(* ========================================================================= *)
(*                                                                           *)
(*                        Complex Matrices                                   *)
(*                                                                           *)
(*        (c) Copyright, Umair Siddique, 2012-2015                           *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last update: February 20, 2014                                            *)
(*                                                                           *)
(* ========================================================================= *)


needs "Multivariate/make_complex.ml";;
needs "utils.ml";;



make_overloadable "**" `:A->B->C`;;
make_overloadable "%%" `:A->B->B`;;


overload_interface ("--",`(cmatrix_neg):complex^N^M->complex^N^M`);;
overload_interface ("+",`(cmatrix_add):complex^N^M->complex^N^M->complex^N^M`);;
overload_interface ("-",`(cmatrix_sub):complex^N^M->complex^N^M->complex^N^M`);;
overload_interface ("%%",`cmatrix_smul:complex->complex^N^M->complex^N^M`);;


overload_interface ("**",`
(cvector_matrix_mul):complex^M->complex^N^M->complex^N`);;

overload_interface ("**",`
(cmatrix_mul):complex^N^M->complex^P^N->complex^P^M`);;

overload_interface ("**",`
(cmatrix_vector_mul):complex^N^M->complex^N->complex^M`);;

parse_as_infix("%%",(21,"right"));;

prioritize_complex();;


let cmatrix_smul = new_definition
  `((%%):complex->complex^N^M->complex^N^M) c A = lambda i j. c * A$i$j`;;

let cmatrix_neg = new_definition
  `!A:complex^N^M. --A = lambda i j. --(A$i$j)`;;

let cmatrix_add = new_definition
  `!A:complex^N^M B:complex^N^M. A + B = lambda i j. A$i$j + B$i$j`;;

let cmatrix_sub = new_definition
  `!A:complex^N^M B:complex^N^M. A - B = lambda i j. A$i$j - B$i$j`;;

let cmatrix_mul = new_definition
  `!A:complex^N^M B:complex^P^N.
        A ** B =
          lambda i j. vsum(1..dimindex(:N)) (\k. A$i$k * B$k$j)`;;

let cmatrix_cvector_mul = new_definition
  `!A:complex^N^M x:complex^N.
        A ** x = lambda i. vsum(1..dimindex(:N)) (\j. A$i$j * x$j)`;;

let cvector_cmatrix_mul = new_definition
  `!A:complex^N^M x:complex^M.
        x ** A = lambda j. vsum(1..dimindex(:M)) (\i. A$i$j * x$i)`;;

let COMPLEX_CART_EQ = prove
  (`!x y:complex^N. x = y <=> !i. 1 <= i /\ i <= dimindex (:N) ==> x$i = y$i`,
  REWRITE_TAC[CART_EQ]);;



let cmat2x2 = new_definition
  `cmat2x2 A B C D :complex^2^2 = vector[vector[A;B]; vector[C;D]]`;;
 
let cmat1 = new_definition
  `cmat1 (k:num) = lambda i j. if i = j then (Cx(&k)) else Cx(&0)`;;




let CCOMMON_TAC xs =
  SIMP_TAC (xs @ [cmat2x2;cmatrix_cvector_mul;COMPLEX_CART_EQ;LAMBDA_BETA;
           CART_EQ;VECTOR_2;DIMINDEX_2;FORALL_2;DOT_2;VSUM_2]);;


let CMAT_2X2_COMPONENTS = prove (`
!a b c d. (cmat2x2 a b c d)$1$1 = a /\ 
          (cmat2x2 a b c d)$1$2 = b /\ 
          (cmat2x2 a b c d)$2$1 = c /\ 
          (cmat2x2 a b c d)$2$2 = d `,

CCOMMON_TAC[]);;


let CMAT2X2_VECTOR_MUL = prove
  (`!A B C D X Y:complex.
    cmat2x2 A B C D ** vector [X;Y] = vector[A*X+B*Y;C*X+D*Y]`,

CCOMMON_TAC[]);;


let CMAT2X2_MUL_VEC = prove 
   (`!A B C D X (Y:complex).
    cmat2x2 A B C D ** vector [X;Y] = vector[A*X+B*Y;C*X+D*Y]`,

CCOMMON_TAC[]);;


let CMAT2X2_VECTOR_MUL_GEN = prove
  (`!(M:complex^2^2) (X:complex) (Y:complex).
    M ** vector [X;Y] = vector[(M$1$1)*X+ (M$1$2)*Y;(M$2$1*X)+ (M$2$2)*Y]`,
  CCOMMON_TAC []);;


let CMAT2X2_EQ = prove
  (`!A B C D P Q R S:complex. cmat2x2 A B C D = cmat2x2 P Q R S <=>
    A = P /\ B = Q /\ C = R /\ D = S`,
  CCOMMON_TAC [] THEN CONV_TAC TAUT);;  


let CVECTOR2_EQ = prove
  (`!x y z t:complex. vector [x;y] :complex^2 = vector [z;t] <=> x=z /\ y=t`,
  CCOMMON_TAC[]);;


let CMAT2X2_MUL = prove
  (`!A B C D P Q R S:complex.
    cmat2x2 A B C D ** cmat2x2 P Q R S =
    cmat2x2 (A*P + B*R) (A*Q + B*S) (C*P + D*R) (C*Q + D*S)`,

CCOMMON_TAC[cmatrix_mul;VECTOR_2;VECTOR_1;LAMBDA_BETA]);;


let CMAT2X2_CMAT1 = prove
  (`cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(&0)) (Cx(&1)) = (cmat1 1)`,

CCOMMON_TAC[cmat1;LAMBDA_BETA] THEN CONV_TAC NUM_REDUCE_CONV);;



let CMAT2X2_RID = prove
  (`!m:complex^2^2. cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(&0)) (Cx(&1)) ** m = m`,
  
CCOMMON_TAC[cmatrix_mul;VECTOR_2;VECTOR_1;LAMBDA_BETA;COMPLEX_MUL_LID;
COMPLEX_MUL_LZERO;COMPLEX_ADD_LID;COMPLEX_ADD_RID]);;


let CMAT2X2_LID = prove
  (`!m:complex^2^2. m ** cmat2x2 (Cx(&1)) (Cx(&0)) (Cx(&0)) (Cx(&1)) = m`,
  
CCOMMON_TAC[cmatrix_mul;VECTOR_2;VECTOR_1;LAMBDA_BETA;COMPLEX_MUL_LID;
            COMPLEX_MUL_RID;COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO;COMPLEX_ADD_LID;
           COMPLEX_ADD_RID]);;


let CMAT2X2_ID = CONJ CMAT2X2_LID CMAT2X2_RID;;


let vector_map = new_definition
  `vector_map (f:A->B) (v:A^N) :B^N = lambda i. f(v$i)`;;

make_overloadable "%" `:A-> B-> B`;; 
 overload_interface("%",`(cvector_mul):complex->complex^N->complex^N`);;

let cvector_mul = new_definition
  `((%):complex->complex^N->complex^N) c x = lambda i. c * x$i`;;


let CMATRIX_SMUL_COMPONENT = prove
 (`!c A:complex^N^M i. (c %% A)$i$j = c * A$i$j`,
  
  REPEAT GEN_TAC THEN
  SUBGOAL_THEN `?k. 1 <= k /\ k <= dimindex(:M) /\ !A:complex^N^M. A$i = A$k`
  CHOOSE_TAC THENL [REWRITE_TAC[FINITE_INDEX_INRANGE]; ALL_TAC] THEN
  SUBGOAL_THEN `?l. 1 <= l /\ l <= dimindex(:N) /\ !z:complex^N. z$j = z$l`
  CHOOSE_TAC THENL [REWRITE_TAC[FINITE_INDEX_INRANGE]; ALL_TAC] THEN
  ASM_SIMP_TAC[cvector_mul;o_DEF;cmatrix_smul;vector_map; COMPLEX_CART_EQ; 
  LAMBDA_BETA]);;


let CMAT2X2_SCALAR_LMUL = prove
  (`!A B C D k. k %% cmat2x2 A B C D = cmat2x2 (k*A) (k*B) (k*C) (k*D)`,
  CCOMMON_TAC [CMATRIX_SMUL_COMPONENT]);;


let CMAT2X2_SCALAR_MUL_ASSOC = prove
  (`!A B C D L M N P k:complex.
    cmat2x2 L M N P ** (k %% cmat2x2 A B C D) = 
    k %% (cmat2x2 L M N P ** cmat2x2 A B C D)`,
    SIMP_TAC[CMAT2X2_SCALAR_LMUL;CMAT2X2_MUL;CMAT2X2_EQ] THEN 
    CONV_TAC COMPLEX_FIELD);;

let CMAT2X2_EQ_SCALAR_LMUL = prove
  (`!A B C D P Q R S. ~(k = Cx(&0)) ==>
    (k %% cmat2x2 A B C D = k %% cmat2x2 P Q R S
      <=> A = P /\ B = Q /\ C = R /\ D = S)`,
   SIMP_TAC[CMAT2X2_SCALAR_LMUL;CMAT2X2_MUL;CMAT2X2_EQ] THEN 
   CONV_TAC COMPLEX_FIELD);;



let CMAT2X2_SCALAR_LMUL_EQ = prove
  (`!A B C D P Q R S. ~(k = Cx(&0)) ==>
    (k %% cmat2x2 A B C D = k %% cmat2x2 P Q R S
      <=> cmat2x2 A B C D =  cmat2x2 P Q R S)`,
   SIMP_TAC[CMAT2X2_SCALAR_LMUL;CMAT2X2_MUL;CMAT2X2_EQ] THEN 
   CONV_TAC COMPLEX_FIELD);;



let FORALL_CMATRIX_2X2 = prove
  (`!P. (!M:complex^2^2. P M) <=> ! a b c d.P (cmat2x2 a b c d )`,
 REWRITE_TAC[FORALL_VECTOR_2;cmat2x2] );;



let CMAT_2X2_EQ_COMPONENT = prove(`
    !(M:complex^2^2). M = cmat2x2 (M$1$1) (M$1$2) (M$2$1) (M$2$2)`,
     CCOMMON_TAC[]);;


let CMAT2X2_VECTOR_MUL_ALT = prove 
 (`!(M:complex^2^2) (x:complex) y x' y'.  (vector[x';y'] = M ** vector[x;y]) = 
    (x' =  (M ** vector[x;y])$1 /\ y' = (M ** vector[x;y])$2)`,
CCOMMON_TAC [ FORALL_CMATRIX_2X2;CVECTOR2_EQ;CMAT2X2_VECTOR_MUL]);;


let CMAT2_MUL = prove (
 ` !(M:complex^2^2). M**M =  cmat2x2((M$1$1 * M$1$1) +  (M$1$2 * M$2$1))
                                ((M$1$1 * M$1$2) +  (M$1$2 * M$2$2))
                                ((M$2$1 * M$1$1) +  (M$2$2 * M$2$1))
                                ((M$2$1 * M$1$2) +  (M$2$2 * M$2$2))`,
 CCOMMON_TAC[cmatrix_mul;LAMBDA_BETA]);;


let CMAT2X2_VECTOR_MUL_ALT2 = prove 
 (`!(M:complex^2^2) (x:complex) y . M ** vector[x;y] =  
     vector[(M$1$1) * x + (M$1$2) * y ; (M$2$1) * x + (M$2$2)* y] `,
CCOMMON_TAC [ FORALL_CMATRIX_2X2;CVECTOR2_EQ;CMAT2X2_VECTOR_MUL]);;


let VSUM_CMUL = prove
  (`!a f s. FINITE s ==> vsum s (\x. a * f x) = a * vsum s f`,
  ONCE_REWRITE_TAC[TAUT `(A ==> B) <=> (A ==> A ==> B)`]
  THEN REPEAT (MATCH_MP_TAC FINITE_INDUCT ORELSE GEN_TAC)
  THEN SIMP_TAC[COMPLEX_VEC_0;COMPLEX_MUL_RZERO;VSUM_CLAUSES;FINITE_INSERT]
  THEN REPEAT GEN_TAC THEN COND_CASES_TAC
  THEN REWRITE_TAC[COMPLEX_ADD_LDISTRIB]);;


let CMATRIX_VECTOR_MUL_ASSOC = prove
  (`!m1:complex^2^2 m2:complex^2^2 x:complex^2.
   m1 ** m2 ** x = (m1 ** m2) ** x`,

  SIMP_TAC[cmatrix_mul;cmatrix_cvector_mul;COMPLEX_CART_EQ;LAMBDA_BETA]
  THEN IMP_REWRITE_TAC[GSYM VSUM_CMUL;FINITE_NUMSEG]
  THEN ONCE_REWRITE_TAC[VSUM_SWAP_NUMSEG]
  THEN IMP_REWRITE_TAC[ONCE_REWRITE_RULE[COMPLEX_MUL_SYM] (GSYM VSUM_CMUL);
    FINITE_NUMSEG]
  THEN REWRITE_TAC[COMPLEX_MUL_AC]);;

