(* ========================================================================= *)
(*                                                                           *)
(*                        Helmholtz Equation &                               *)
(*                        Gaussian Beam Verification                         *)
(*                                                                           *)
(*        (c) Copyright, Umair Siddique, 2012-2015                           *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <muh_sidd@ece.concordia.ca>                                  *)
(*                                                                           *)
(* Last updated:  Dec 2015                                                   *)
(*                                                                           *)
(* ========================================================================= *)


needs "q_parameters_beams.ml";;


let laplacian = new_definition `
    laplacian f (x,y) = 
    higher_complex_derivative 2  (\x. f (x,y)) x +  
    higher_complex_derivative 2  (\y. f (x,y)) y `;;


let Paraxial_Helmholtz_eq = define `
    Paraxial_Helmholtz_eq U (x,y,z) k <=>
    (laplacian (\ (x,y). U (x,y,z)) (x,y)) - 
    (ii* Cx(&2) * k * complex_derivative (\z. U (x,y,z)) z) = Cx(&0)`;;


let paraxial_wave = new_definition `
    paraxial_wave A x y z k = A(x,y,z) * cexp (--ii*k* Cx z)`;;


let paraboloid_wave_sol = new_definition ` 
    paraboloid_wave_sol A0 z zr x y k =
    (A0 /(z + ii * zr ))*
    cexp(--ii*k*((x pow 2 + y pow 2)/(Cx(&2)*(z + ii*zr))))`;;

let q_parameter_amplitude = new_definition ` 
    q_parameter_amplitude A0 z x y k w0 lam =
    (A0 /qq (z,w0,lam))*
    cexp(--ii*k*((x pow 2 + y pow 2)/(Cx(&2)*(qq (z,w0,lam)))))`;;

(*===
let gaussian_beam_qq = new_definition ` 
    gaussian_beam_qq A0 z x y k w0 lam =
    (A0 /qq (z,w0,lam))*
    cexp(--ii*k*((x pow 2 + y pow 2)/(Cx(&2)*(qq (z,w0,lam)))))`;;



 ==   **)

    
let gaussian_beam_intensity = new_definition `
    gaussian_beam_intensity  A x y z k  = 
       (norm (paraxial_wave A x y z k)) pow 2`;;


(*==================================================================*)
(*                  IMPORTANT LEMMAS                                *)
(*==================================================================*)

let VSUM_22 = prove
 (`!t. vsum(0..2) t = t(0) + t(1) + t(2)`,
  REWRITE_TAC[num_CONV `1`; num_CONV `2`; VSUM_CLAUSES_NUMSEG] THEN
  REWRITE_TAC[VSUM_SING_NUMSEG; ARITH; VECTOR_ADD_ASSOC]);;


let HIGHER_COMPLEX_DERIVATIVE_MUL_2 = prove
 (`! f g s z.  
    open s /\ f holomorphic_on s /\ g holomorphic_on s /\ z IN s
    ==> higher_complex_derivative 2 (\w. f w * g w) z = 
        f z * higher_complex_derivative 2 g z +
        Cx (&2) * complex_derivative f z * 
                  complex_derivative g z +
        higher_complex_derivative 2 f z * g z`,
  REPEAT STRIP_TAC THEN IMP_REWRITE_TAC [HIGHER_COMPLEX_DERIVATIVE_MUL] THEN
  SIMP_TAC [VSUM_22] THEN SIMP_TAC [binom; BINOM_1] THEN
  SIMP_TAC [ARITH_RULE `(2 - 0 = 2) /\ (1 - 0) = 1 /\ 2 -1 = 1 /\ 2 - 2 = 0`] THEN
  SIMP_TAC
    [COMPLEX_MUL_LID; higher_complex_derivative; HIGHER_COMPLEX_DERIVATIVE_1;
     BINOM_REFL]);;



let HIGHER_COMPLEX_DERIVATIVE_CMUL = prove
 (`!n g a s. (!x.(open s) /\  x IN s /\
                     g holomorphic_on s) ==>
   higher_complex_derivative n (\x. a * g x) x = 
   a * (higher_complex_derivative n g x)`,
  INDUCT_TAC THENL [SIMP_TAC [higher_complex_derivative_alt]; ALL_TAC] THEN
  ASM_SIMP_TAC [higher_complex_derivative_alt] THEN
  SIMP_TAC [COMPLEX_DERIVATIVE_LMUL] THEN REPEAT STRIP_TAC THEN
  ASSERT_TAC
    `! w  c. complex_derivative (\w. c * g w) = 
  (\x. c * complex_derivative g x)` THENL
   [REPEAT STRIP_TAC THEN REWRITE_TAC [FUN_EQ_THM] THEN GEN_TAC THEN
    MATCH_MP_TAC COMPLEX_DERIVATIVE_LMUL THEN
    MATCH_MP_TAC HOLOMORPHIC_ON_IMP_DIFFERENTIABLE_AT THEN
    EXISTS_TAC `(s:real^2->bool)` THEN ASM_SIMP_TAC [];
    ALL_TAC] THEN
  ASM_SIMP_TAC [COMPLEX_DERIVATIVE_LMUL] THEN ASM_SIMP_TAC [] THEN
  UNDISCH_TAC
    `!g a s.
          ( !x. open s /\ x IN s /\ g holomorphic_on s)
          ==> higher_complex_derivative n (\x. a * g x) x =
              a * higher_complex_derivative n g x` THEN
  SIMP_TAC [RIGHT_FORALL_IMP_THM] THEN
  DISCH_THEN (MP_TAC o SPEC `(\x. complex_derivative  g x)`) THEN
  DISCH_THEN (MP_TAC o SPEC `a:complex`) THEN
  DISCH_THEN (MP_TAC o SPEC `(s:real^2->bool)`) THEN ANTS_TAC THENL
   [ALL_TAC; SIMP_TAC [ETA_AX]] THEN
  ASM_SIMP_TAC [] THEN
  MP_TAC
    (ISPECL [`(g:real^2->real^2)`; `(s:real^2->bool)`; `(1:num)`]
       HOLOMORPHIC_HIGHER_COMPLEX_DERIVATIVE) THEN
  SIMP_TAC [HIGHER_COMPLEX_DERIVATIVE_1] THEN ASM_SIMP_TAC [] THEN
  ASSERT_TAC `(\x. complex_derivative g x) = higher_complex_derivative 1 g` THENL
   [SIMP_TAC [FUN_EQ_THM; HIGHER_COMPLEX_DERIVATIVE_1]; ASM_MESON_TAC []]);;


let HOLOMORPHIC_GAUSSIAN_BEAM = prove
 (`!s A0 z zr y z k. 
   (\x. paraboloid_wave_sol A0 z zr x y k)  holomorphic_on s`,
  REPEAT STRIP_TAC THEN REWRITE_TAC [paraboloid_wave_sol] THEN
  MATCH_MP_TAC HOLOMORPHIC_ON_LMUL THEN REWRITE_TAC [holomorphic_on] THEN
  REPEAT STRIP_TAC THEN
  EXISTS_TAC
    `(--ii *
  k *
  (Cx (&2) * x pow (2 - 1) * Cx (&1) + Cx (&0)) / (Cx (&2) * (z + ii * zr))) *
 cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr)))` THEN
  COMPLEX_DIFF_TAC THEN SIMP_TAC []);;



let WAVE_SOL_LEMMA_1 = prove
 (`!A0 zr k y z zr x. 
   complex_derivative (\x. paraboloid_wave_sol A0 z zr x y k) x = 
   A0 / (z + ii * zr) *
   (--ii * k * x / (z + ii * zr)) *
   cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr)))`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE THEN
  REWRITE_TAC [paraboloid_wave_sol] THEN COMPLEX_DIFF_TAC THEN
  SIMP_TAC
    [COMPLEX_ADD_RID; COMPLEX_MUL_RID; ARITH_RULE `2 - 1 = 1`; COMPLEX_POW_1] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `! a b  d. ~ (a = Cx(&0)) ==> (a*b) / (a*d) = b / d`] THEN
  SIMP_TAC [CX_INJ; REAL_FIELD `~(&2 = &0)`]);;


let WAVE_SOL_LEMMA_1_ALT = prove
 (`!A0 zr k y z. 
   complex_derivative (\x. paraboloid_wave_sol A0 z zr x y k) = 
   (\x. A0 / (z + ii * zr) *
   (--ii * k * x / (z + ii * zr)) *
   cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr))))`,
  REWRITE_TAC [FUN_EQ_THM] THEN MESON_TAC [WAVE_SOL_LEMMA_1]);;



let WAVE_SOL_LEMMA_2 = prove
 (`!A0 zr k y z zr x. 
   complex_derivative (\x.
   A0 / (z + ii * zr) *
         (--ii * k * x / (z + ii * zr)) *
         cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr)))) x =
A0 / (z + ii * zr) *
 ((--ii * k * x / (z + ii * zr)) *
  (--ii * k * x / (z + ii * zr)) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )) +
  (--ii * k * Cx (&1) / (z + ii * zr)) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )))`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE THEN
  COMPLEX_DIFF_TAC THEN
  SIMP_TAC
    [COMPLEX_ADD_RID; COMPLEX_MUL_RID; ARITH_RULE `2 - 1 = 1`; COMPLEX_POW_1] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `! a b  d. ~ (a = Cx(&0)) ==> (a*b) / (a*d) = b / d`] THEN
  SIMP_TAC [CX_INJ; REAL_FIELD `~(&2 = &0)`]);;


let WAVE_SOL_DERV_X = prove
 (`!A0 zr k y z x zr. higher_complex_derivative 2
  (\x. paraboloid_wave_sol A0 z zr x y k) x = 
 A0 / (z + ii * zr) *
 ((--ii * k * x / (z + ii * zr)) *
  (--ii * k * x / (z + ii * zr)) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )) +
  (--ii * k * Cx (&1) / (z + ii * zr)) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )))`,
  SIMP_TAC [ARITH_RULE `2 = SUC 1`] THEN
  REWRITE_TAC [higher_complex_derivative_alt; HIGHER_COMPLEX_DERIVATIVE_1] THEN
  REWRITE_TAC [WAVE_SOL_LEMMA_1_ALT] THEN REWRITE_TAC [WAVE_SOL_LEMMA_2] THEN
  SIMP_TAC [ARITH_RULE `2 = SUC 1`]);;


let WAVE_SOL_LEMMA_3 = prove
 (`!A0 zr k y z x zr. 
   complex_derivative (\y. paraboloid_wave_sol A0 z zr x y k) y = 
   A0 / (z + ii * zr) *
   (--ii * k * y / (z + ii * zr)) *
   cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr)))`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE THEN
  REWRITE_TAC [paraboloid_wave_sol] THEN COMPLEX_DIFF_TAC THEN
  SIMP_TAC
    [COMPLEX_ADD_RID; COMPLEX_MUL_RID; COMPLEX_ADD_LID;
     ARITH_RULE `2 - 1 = 1`; COMPLEX_POW_1] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `! a b  d. ~ (a = Cx(&0)) ==> (a*b) / (a*d) = b / d`] THEN
  SIMP_TAC [CX_INJ; REAL_FIELD `~(&2 = &0)`]);;


let WAVE_SOL_LEMMA_3_ALT = prove
 (`!A0 zr k y z x. 
   complex_derivative (\y. paraboloid_wave_sol A0 z zr x y k)  = 
   (\y. A0 / (z + ii * zr) *
   (--ii * k * y / (z + ii * zr)) *
   cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr))))`,
  REWRITE_TAC [FUN_EQ_THM] THEN MESON_TAC [WAVE_SOL_LEMMA_3]);;



let WAVE_SOL_LEMMA_4 = prove
 (`!A0 zr k y z x y  zr. 
   complex_derivative (\y. A0 / (z + ii * zr) *
              (--ii * k * y / (z + ii * zr)) *
              cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) ))) y = 
A0 / (z + ii * zr) *
 ((--ii * k * y / (z + ii * zr)) *
  (--ii * k * y / (z + ii * zr)) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )) +
  (--ii * k * Cx (&1) /(z + ii * zr) ) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )))`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE THEN
  COMPLEX_DIFF_TAC THEN
  SIMP_TAC
    [COMPLEX_ADD_RID; COMPLEX_ADD_LID; COMPLEX_MUL_RID;
     ARITH_RULE `2 - 1 = 1`; COMPLEX_POW_1] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `! a b  d. ~ (a = Cx(&0)) ==> (a*b) / (a*d) = b / d`] THEN
  SIMP_TAC [CX_INJ; REAL_FIELD `~(&2 = &0)`]);;


let WAVE_SOL_DERV_Y = prove
 (`!A0 zr k y z x. higher_complex_derivative 2
  (\y. paraboloid_wave_sol A0 z zr x y k) y = 
  A0 / (z + ii * zr) *
 ((--ii * k * y / (z + ii * zr)) *
  (--ii * k * y / (z + ii * zr)) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )) +
  (--ii * k * Cx (&1) /(z + ii * zr) ) *
  cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) *(z + ii * zr) )))`,
  SIMP_TAC [ARITH_RULE `2 = SUC 1`] THEN
  REWRITE_TAC [higher_complex_derivative_alt; HIGHER_COMPLEX_DERIVATIVE_1] THEN
  REWRITE_TAC [WAVE_SOL_LEMMA_3_ALT] THEN REWRITE_TAC [WAVE_SOL_LEMMA_4] THEN
  SIMP_TAC [ARITH_RULE `2 = SUC 1`]);;



let WAVE_SOL_DERV_Z = prove
 (`!A0 zr k y z x. (~(z + ii * zr = Cx (&0)) /\ ~(Cx (&2) * (z + ii * zr)  = Cx (&0))) ==>
   complex_derivative (\z. paraboloid_wave_sol A0 z zr x y k) z = 
   A0 / (z + ii * zr) *
   (ii * k * ((x pow 2 + y pow 2) * Cx (&2)) / (Cx (&2) * (z + ii * zr)) pow 2) *
    cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr))) +
    --A0 / (z + ii * zr) pow 2 *
    cexp (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * (z + ii * zr)))`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE THEN
  REWRITE_TAC [paraboloid_wave_sol] THEN COMPLEX_DIFF_TAC THEN
  ASM_SIMP_TAC [] THEN
  SIMP_TAC
    [COMPLEX_MUL_LZERO; COMPLEX_MUL_LID; COMPLEX_ADD_RID; COMPLEX_MUL_RZERO;
     COMPLEX_MUL_RID; COMPLEX_ADD_LID; ARITH_RULE `2 - 1 = 1`;
     COMPLEX_FIELD `!a. Cx (&0) - a = --a`;
     COMPLEX_FIELD `!a. a - Cx (&0)  = a`;
     COMPLEX_FIELD `!(a:complex) b c d. -- a*b *(--c/d) = a*b*c/d`;
     COMPLEX_POW_1]);;



let HELMHOLTZ_EQUATION_LEMMA  = prove
 (`!A0 z zr x y k. 
   (~(z + ii * zr = Cx (&0)) /\ ~(Cx (&2) * (z + ii * zr)  = Cx (&0)))
    ==> (Paraxial_Helmholtz_eq (\ (x,y,z). 
         paraboloid_wave_sol A0 z zr x y k) (x,y,z) k)`,
  REWRITE_TAC [Paraxial_Helmholtz_eq; laplacian] THEN REPEAT STRIP_TAC THEN
  REWRITE_TAC [WAVE_SOL_DERV_Z; WAVE_SOL_DERV_X; WAVE_SOL_DERV_Y] THEN
  ASM_SIMP_TAC [WAVE_SOL_DERV_Z] THEN
  SIMP_TAC [COMPLEX_MUL_ASSOC; GSYM COMPLEX_ADD_RDISTRIB] THEN
  CONV_TAC COMPLEX_FIELD);;


let HELMHOLTZ_EQUATION_VERIFIED  = prove
 (`!A0 z zr x y k. ~(z + ii * zr = Cx (&0))
    ==> (Paraxial_Helmholtz_eq (\ (x,y,z). 
         paraboloid_wave_sol A0 z zr x y k) (x,y,z) k)`,
  SIMP_TAC
    [COMPLEX_FIELD
       `~(z + ii * zr = Cx (&0)) = 
                ~(Cx (&2) * (z + ii * zr) = Cx (&0))`;
     HELMHOLTZ_EQUATION_LEMMA]);;


let HELMHOLTZ_EQUATION_VERIFIED_Q_PARAMETERS_LEMMA = prove
 (` !A0 x y z w0 lam k.
   ~(qq (z,w0,lam) = Cx (&0)) ==> 
     Paraxial_Helmholtz_eq (\(x,y,z). 
     paraboloid_wave_sol A0 z (Cx (rayleigh_range w0 lam)) x y k)
     (x,y,Cx z) k`, SIMP_TAC[qq;HELMHOLTZ_EQUATION_VERIFIED]);;




let HELMHOLTZ_EQUATION_VERIFIED_Q_PARAMETERS_ALT = prove
 (`!A0 x y z w0 lam k.
   &0 < w0 /\ &0 < lam  ==> 
     Paraxial_Helmholtz_eq (\(x,y,z). 
     paraboloid_wave_sol A0 z (Cx (rayleigh_range w0 lam)) x y k)(x,y,Cx z) k`,
  SIMP_TAC [HELMHOLTZ_EQUATION_VERIFIED_Q_PARAMETERS_LEMMA; Q_NZ]);;


let GAUSSIAN_BEAM_EQ_WAVE_SOL = prove
 (`!A0 x y z w0 lam k. 
   q_parameter_amplitude A0 z x y k w0 lam = 
   paraboloid_wave_sol A0 (Cx(z)) (Cx (rayleigh_range w0 lam)) x y k`,
  SIMP_TAC [paraboloid_wave_sol;q_parameter_amplitude; qq]);;


let COMPLEX_IM_LEMMA = prove
 (`!z. Im z = &0 ==>  Cx (Re z) = z`,
  SIMP_TAC [COMPLEX_EXPAND] THEN ONCE_REWRITE_TAC [COMPLEX_EXPAND] THEN
  SIMP_TAC [GSYM COMPLEX_TRAD] THEN SIMP_TAC [GSYM COMPLEX_TRAD; IM] THEN
  SIMP_TAC [RE_CX] THEN SIMP_TAC [COMPLEX_EQ] THEN
  SIMP_TAC [COMPLEX_EQ; RE; IM] THEN SIMP_TAC [IM_CX]);;


let LCM_LEMMA_2 = 
    REAL_FIELD `! a b c.  ~(c = &0) ==> a + b / c = (a*c + b)/c`;;

let LCM_LEMMA_3 = 
    REAL_FIELD `! a b .  ~(a = &0) ==> a pow 2 + b pow 2  =  
                           a pow 2 *(&1 + b pow 2 / a pow 2)`;;

let LCM_LEMMA_4 = 
    REAL_FIELD `! a b .  ~(b = &0) ==> a pow 2 + b pow 2  =  
                           b pow 2 *(&1 + a pow 2 / b pow 2)`;;


let PARAXIAL_WAVE_ALT_LEMMA = prove
 (`! w0 lam z. &0 < w0 /\ &0 < lam /\ ~(z = &0) ==>
(Cx (&1) / Cx (RR z w0 lam) -
   ii * Cx lam / (Cx pi * Cx (WW z w0 lam pow 2))) = 
  Cx (&1) /
 (ii *
  Cx (rayleigh_range w0 lam) *
  Cx (sqrt (&1 + z pow 2 / rayleigh_range w0 lam pow 2))) *
 (ii * Cx z + Cx (rayleigh_range w0 lam)) /
 Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))`,
  REPEAT STRIP_TAC THEN SIMP_TAC [RR_ALTERNATIVE_FORM] THEN
  REWRITE_TAC [RR; WW] THEN SIMP_TAC [REAL_POW_DIV] THEN
  SIMP_TAC [RR_ALTERNATIVE_FORM] THEN SIMP_TAC [REAL_POW_MUL] THEN
  IMP_REWRITE_TAC [HALFPOWER_SQUARE_LEMMA] THEN
  IMP_REWRITE_TAC [ONE_PLUS_X_POS] THEN REWRITE_TAC [GSYM REAL_POW_DIV] THEN
  IMP_REWRITE_TAC [SQUARE_POS] THEN
  IMP_REWRITE_TAC
    [RAYLEIGH_RANGE_NZ_2;
     REAL_FIELD `~(x= &0)/\  ~(y = &0) ==> ~(x / y = &0)`] THEN
  REWRITE_TAC [REAL_POW_DIV] THEN
  IMP_REWRITE_TAC [RAYLEIGH_RANGE_NZ; LCM_LEMMA_2] THEN
  SIMP_TAC [REAL_MUL_LID] THEN SIMP_TAC [GSYM CX_MUL] THEN
  SIMP_TAC
    [complex_div; COMPLEX_INV_MUL; GSYM CX_INV; COMPLEX_INV_INV;
     COMPLEX_MUL_LID; COMPLEX_INV_DIV; CX_DIV; REAL_INV_DIV; CX_MUL] THEN
  SIMP_TAC [CX_INV; CX_MUL; complex_div; CX_DIV] THEN
  SIMP_TAC [COMPLEX_INV_MUL; COMPLEX_INV_INV] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `Cx z * inv (Cx (z pow 2 + rayleigh_range w0 lam pow 2)) -
 ii *
 Cx lam *
 inv (Cx pi) *
 inv (Cx (w0 pow 2)) *
 inv (Cx (rayleigh_range w0 lam pow 2 + z pow 2)) *
 Cx (rayleigh_range w0 lam pow 2) = 
 Cx z / (Cx (z pow 2 + rayleigh_range w0 lam pow 2)) -
 (ii * Cx (rayleigh_range w0 lam pow 2)* Cx lam * inv ((Cx pi) * Cx (w0 pow 2))) /
 Cx (rayleigh_range w0 lam pow 2 + z pow 2)`] THEN
  SIMP_TAC [REAL_ADD_SYM; COMPLEX_FIELD `a / b - c / b = (a - c) / b`] THEN
  REWRITE_TAC
    [MESON [REAL_POW_2; CX_MUL]
       `Cx(rayleigh_range w0 lam pow 2) = 
                                  Cx(rayleigh_range w0 lam)  *
                                  Cx(rayleigh_range w0 lam)`] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `ii *
  (Cx (rayleigh_range w0 lam) * Cx (rayleigh_range w0 lam)) *
  Cx lam *
  inv (Cx pi * Cx (w0 pow 2)) = (Cx (rayleigh_range w0 lam) *  Cx lam * inv (Cx pi * Cx (w0 pow 2)))*ii *
  (Cx (rayleigh_range w0 lam))`] THEN
  ASSERT_TAC
    `(Cx (rayleigh_range w0 lam) * Cx lam * inv (Cx pi * Cx (w0 pow 2))) = Cx(&1)` THENL
   [REWRITE_TAC [rayleigh_range] THEN
    REWRITE_TAC [GSYM CX_INV; GSYM CX_DIV; GSYM CX_MUL; CX_INJ] THEN
    REWRITE_TAC [GSYM real_div] THEN MP_TAC PI_POS THEN
    REPEAT (POP_ASSUM MP_TAC) THEN CONV_TAC REAL_FIELD;
    ALL_TAC] THEN
  ASM_SIMP_TAC [COMPLEX_MUL_LID; COMPLEX_MUL_RID] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `!z w0 lam m. (Cx z - ii * Cx (rayleigh_range w0 lam))/m = 
                (ii*(Cx z - ii * Cx (rayleigh_range w0 lam)))/ (ii*m)`] THEN
  SIMP_TAC
    [COMPLEX_SUB_LDISTRIB; COMPLEX_FIELD `! a b.  a - ii*ii*b = a + b`] THEN
  ASSERT_TAC
    `Cx(z pow 2 + rayleigh_range w0 lam pow 2) = 
              Cx(sqrt (z pow 2 + rayleigh_range w0 lam pow 2) *
              sqrt (z pow 2 + rayleigh_range w0 lam pow 2))` THENL
   [SIMP_TAC [CX_INJ; GSYM REAL_POW_2] THEN
    IMP_REWRITE_TAC
      [SQRT_POW_2; REAL_LT_IMP_LE;
       REAL_FIELD `!a b.
 &0 < a /\ &0 < b ==> &0 < a + b`] THEN
    ASM_SIMP_TAC [REAL_LT_POW_2; RAYLEIGH_RANGE_NZ_2];
    ALL_TAC] THEN
  ASM_SIMP_TAC [] THEN SIMP_TAC [CX_MUL] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(ii * Cx z + Cx (rayleigh_range w0 lam)) /
 (ii *
  Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2)) *
  Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))) = Cx(&1) / (ii *
  Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))) * (ii * Cx z + Cx (rayleigh_range w0 lam)) / Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))`] THEN
  ASSERT_TAC
    `(ii * Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))) = 
(ii*Cx(rayleigh_range w0 lam)* Cx(sqrt(&1 + z pow 2 /rayleigh_range w0 lam pow 2)))` THENL
   [IMP_REWRITE_TAC [LCM_LEMMA_4; RAYLEIGH_RANGE_NZ_2] THEN
    SIMP_TAC [SQRT_MUL] THEN
    IMP_REWRITE_TAC [POW_2_SQRT; REAL_LT_IMP_LE; RAYLEIGH_POS_2] THEN
    SIMP_TAC [CX_MUL];
    ASM_SIMP_TAC [] THEN SIMP_TAC [GSYM COMPLEX_MUL_ASSOC] THEN
    IMP_REWRITE_TAC [LCM_LEMMA_2; RAYLEIGH_RANGE_NZ] THEN
    SIMP_TAC [REAL_MUL_LID] THEN
    SIMP_TAC [complex_div; COMPLEX_INV_MUL; COMPLEX_MUL_LID; CX_DIV; CX_MUL] THEN
    SIMP_TAC [GSYM COMPLEX_MUL_ASSOC] THEN MESON_TAC [REAL_ADD_SYM]]);;




let ATN_CEXP_LEMMA = prove
 (`! z w0 lam. &0 < w0 /\ &0 < lam ==>
cexp (ii* Cx(atn (z / (rayleigh_range w0 lam)))) = 
 (ii * Cx z + Cx (rayleigh_range w0 lam)) /
  Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))`,
  REPEAT STRIP_TAC THEN SIMP_TAC [CEXP_EULER; GSYM CX_COS; GSYM CX_SIN] THEN
  SIMP_TAC [COS_ATN; SIN_ATN] THEN SIMP_TAC [REAL_POW_DIV] THEN
  IMP_REWRITE_TAC [LCM_LEMMA_2] THEN ASM_SIMP_TAC [RAYLEIGH_RANGE_NZ] THEN
  SIMP_TAC [REAL_MUL_LID] THEN SIMP_TAC [SQRT_DIV] THEN
  IMP_REWRITE_TAC [POW_2_SQRT; REAL_LT_IMP_LE; RAYLEIGH_POS_2] THEN
  SIMP_TAC
    [CX_DIV; CX_MUL; complex_div; COMPLEX_INV_INV; COMPLEX_INV_MUL;
     COMPLEX_MUL_LID] THEN
  REWRITE_TAC [GSYM complex_div] THEN
  SIMP_TAC [COMPLEX_FIELD `(a + b) / c = a / c + b / c`] THEN
  SIMP_TAC [complex_div] THEN SIMP_TAC [CX_INV] THEN
  SIMP_TAC [GSYM COMPLEX_MUL_ASSOC] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `inv (Cx (sqrt (rayleigh_range w0 lam pow 2 + z pow 2))) *
 Cx (rayleigh_range w0 lam) +
 ii *
 Cx z *
 inv (Cx (rayleigh_range w0 lam)) *
 inv (Cx (sqrt (rayleigh_range w0 lam pow 2 + z pow 2))) *
 Cx (rayleigh_range w0 lam) = inv (Cx (sqrt (rayleigh_range w0 lam pow 2 + z pow 2))) *
 Cx (rayleigh_range w0 lam) +
 ii *
 Cx z *
 inv (Cx (sqrt (rayleigh_range w0 lam pow 2 + z pow 2))) *
 (Cx (rayleigh_range w0 lam) * inv (Cx (rayleigh_range w0 lam)))`] THEN
  IMP_REWRITE_TAC [COMPLEX_MUL_RINV; CX_INJ; RAYLEIGH_RANGE_NZ_2] THEN
  SIMP_TAC [COMPLEX_MUL_RID] THEN SIMP_TAC [REAL_ADD_SYM] THEN
  CONV_TAC COMPLEX_FIELD);;

let gaussian_beam_qq = q_parameter_amplitude ;;


let PARAXIAL_WAVE_SOLUTION_GENERAL_FORM = prove
 (`!z w0 lam A0. &0 < w0 /\ &0 < lam /\ ~(z = &0) ==>
    paraxial_wave ( \ (x,y,z). q_parameter_amplitude A0 z x y k w0 lam) x y z k = 
(A0 / (ii * Cx (rayleigh_range w0 lam)) * Cx w0 / Cx (WWW z w0 lam)) *
cexp
 (--Cx (&1) *
  (k * Cx lam) / (Cx (&2) * Cx pi) *
  (x pow 2 + y pow 2) / Cx (WW z w0 lam pow 2) +
  --ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) +
  --ii * k * Cx z + ii * Cx (atn (z / rayleigh_range w0 lam)))`,
  REPEAT STRIP_TAC THEN REWRITE_TAC [gaussian_beam_qq; paraxial_wave] THEN
  ONCE_SIMP_TAC
    [COMPLEX_FIELD `A0 / qq (z,w0,lam) = A0 * (Cx(&1) / qq(z,w0,lam))`;
     COMPLEX_FIELD
       `--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * qq (z,w0,lam)) = 
           --ii * k * (x pow 2 + y pow 2) / Cx (&2) * (Cx(&1) / qq (z,w0,lam))`] THEN
  ASM_SIMP_TAC [qq_QQ] THEN
  SIMP_TAC
    [QQ; complex_div; COMPLEX_INV_INV; COMPLEX_INV_MUL; COMPLEX_MUL_LID] THEN
  REWRITE_TAC [GSYM COMPLEX_MUL_ASSOC; COMPLEX_SUB_LDISTRIB] THEN
  REWRITE_TAC [GSYM COMPLEX_INV_MUL] THEN REWRITE_TAC [GSYM complex_div] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) -
  --ii *
  k *
  (x pow 2 + y pow 2) *
  inv (Cx (&2)) *
  ii *
  Cx lam / (Cx pi * Cx (WW z w0 lam pow 2))) = (--ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) +
  ii pow 2 *
  ((k*Cx lam) / (Cx (&2)* Cx(pi)))*
  (x pow 2 + y pow 2) / Cx (WW z w0 lam pow 2))`] THEN
  SIMP_TAC [CEXP_ADD] THEN SIMP_TAC [COMPLEX_POW_II_2] THEN
  SIMP_TAC [GSYM COMPLEX_MUL_ASSOC] THEN
  SIMP_TAC [COMPLEX_FIELD `!a b c d e. a*b*c*d*e = (a*b)*d*(c*e)`] THEN
  SIMP_TAC [GSYM CEXP_ADD] THEN
  SIMP_TAC
    [COMPLEX_FIELD `inv (Cx (RR z w0 lam)) = Cx(&1)/ (Cx (RR z w0 lam))`] THEN
  ASM_SIMP_TAC [PARAXIAL_WAVE_ALT_LEMMA] THEN
  ASM_SIMP_TAC [GSYM REAL_POW_DIV; WWW_ALT] THEN SIMP_TAC [CX_DIV] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(A0 *
  Cx (&1) / (ii * Cx (rayleigh_range w0 lam) * Cx (WWW z w0 lam) / Cx w0) *
  (ii * Cx z + Cx (rayleigh_range w0 lam)) /
  Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))) = ((A0 / (ii * Cx (rayleigh_range w0 lam))) * (Cx w0 / Cx (WWW z w0 lam))) *
  (ii * Cx z + Cx (rayleigh_range w0 lam)) /
  Cx (sqrt (z pow 2 + rayleigh_range w0 lam pow 2))`] THEN
  ASM_SIMP_TAC [GSYM ATN_CEXP_LEMMA] THEN
  SIMP_TAC [CEXP_ADD; GSYM COMPLEX_MUL_ASSOC] THEN
  SIMP_TAC [complex_div; COMPLEX_INV_MUL] THEN
  SIMP_TAC [GSYM COMPLEX_MUL_ASSOC] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(--Cx (&1) *
  k *
  inv (Cx (&2)) *
  inv (Cx pi) *
  Cx lam *
  (x pow 2 + y pow 2) *
  inv (Cx (WW z w0 lam pow 2))) = (--Cx (&1) *
  k *
  Cx lam *
  inv (Cx (&2)) *
  inv (Cx pi) *
  (x pow 2 + y pow 2) *
  inv (Cx (WW z w0 lam pow 2)))`] THEN
  CONV_TAC COMPLEX_FIELD);;


let LEMMA_XX = COMPLEX_FIELD `(--Cx (&1) *
              (k * Cx lam) / (Cx (&2) * Cx pi) *
              (x pow 2 + y pow 2) / Cx (WW z w0 lam pow 2) +
              --ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) +
              --ii * k * Cx z +
              ii * Cx (atn (z / rayleigh_range w0 lam))) = --((k * Cx lam) / (Cx (&2) * Cx pi) *
              (x pow 2 + y pow 2) / Cx (WW z w0 lam pow 2)) +
(--ii * k * Cx z -
   ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) +
   ii * Cx (atn (z / rayleigh_range w0 lam)))`;;


let GAUSSIAN_BEAM_GEN_FORM_LEMMA_1 = REWRITE_RULE [LEMMA_XX]
PARAXIAL_WAVE_SOLUTION_GENERAL_FORM ;;

let GAUSSIAN_BEAM_GEN_FORM_LEMMA_1 = ONCE_REWRITE_RULE [CEXP_ADD]
GAUSSIAN_BEAM_GEN_FORM_LEMMA_1 ;;

let GAUSSIAN_BEAM_GEN_FORM_LEMMA_2 = prove(
    `!z w0 lam A0.
         &0 < w0 /\ &0 < lam /\ ~(z = &0)
         ==>
	     paraxial_wave
             (\(x,y,z). q_parameter_amplitude A0 z x y k w0 lam) x y z k =
             let Ac = A0 / (ii * Cx (rayleigh_range w0 lam)) in
             (Ac *
              Cx w0 / Cx (WWW z w0 lam)) *
             cexp
             (--((k * Cx lam) / (Cx (&2) * Cx pi) *
                 (x pow 2 + y pow 2) / Cx (WW z w0 lam pow 2))) *
             cexp
             (--ii * k * Cx z -
              ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) +
              ii * Cx (atn (z / rayleigh_range w0 lam)))`,
REPEAT GEN_TAC THEN LET_TAC THEN ASM_MESON_TAC[ GAUSSIAN_BEAM_GEN_FORM_LEMMA_1]);;



let GAUSSIAN_BEAM_GEN_FORM_LEMMA_2 =
    SIMP_RULE [GSYM WW_WWW] GAUSSIAN_BEAM_GEN_FORM;;


let GAUSSIAN_BEAM_GENERAL_FORM  = prove(
     `!z w0 lam A0.
         &0 < w0 /\ &0 < lam /\ ~(z = &0)
         ==> paraxial_wave
             (\(x,y,z). q_parameter_amplitude A0 z x y k w0 lam) x y z k =
             (let Ac = A0 / (ii * Cx (rayleigh_range w0 lam)) in
              (Ac * Cx w0 / Cx (WW z w0 lam)) *
              cexp
              (--((k * Cx lam) / (Cx (&2) * Cx pi) *
                  (x pow 2 + y pow 2) / Cx (WW z w0 lam pow 2))) *
              cexp
              (--ii * k * Cx z -
               ii * k * (x pow 2 + y pow 2) / (Cx (&2) * Cx (RR z w0 lam)) +
               ii * Cx (atn (z / rayleigh_range w0 lam))))`,

MESON_TAC[GAUSSIAN_BEAM_GEN_FORM_LEMMA_2]);;





let GAUSSIAN_BEAM_INTENSITY = prove
 (`!z w0 lam A0 k. 
 &0 < w0 /\ &0 < lam /\ ~(z = &0) ==> 
 gaussian_beam_intensity (\ (a,b,z).
q_parameter_amplitude  A0 z (a) (b) (Cx k) w0 lam) (Cx x) (Cx y) (z) (Cx k) = 
norm (A0 / (ii * Cx (rayleigh_range w0 lam))) pow 2 *
   abs (w0 / WWW z w0 lam) pow 2 *
   abs (exp (-- &1 * (k * lam) / (&2 * pi) * 
               (x pow 2 + y pow 2) / WW z w0 lam pow 2)) pow 2`,
  REPEAT STRIP_TAC THEN REWRITE_TAC [gaussian_beam_intensity] THEN
  ASM_SIMP_TAC [PARAXIAL_WAVE_SOLUTION_GENERAL_FORM] THEN
  SIMP_TAC [CEXP_ADD; COMPLEX_NORM_MUL; GSYM CX_POW; GSYM CX_ADD] THEN
  REWRITE_TAC [GSYM CX_MUL] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `(--ii * Cx (k * z)) = 
                           (ii * --Cx (k * z))`] THEN
  SIMP_TAC
    [COMPLEX_FIELD `--ii * a = ii * --a`; GSYM CX_DIV; GSYM CX_MUL;
     GSYM CX_NEG; NORM_CEXP_II; REAL_MUL_LID; REAL_MUL_RID] THEN
  SIMP_TAC [GSYM CX_EXP; COMPLEX_NORM_CX] THEN SIMP_TAC [REAL_POW_MUL] THEN
  REAL_ARITH_TAC);;



let GAUSSIAN_BEAM_INTENSITY_ALT_1 = prove(
     `!z w0 lam A0 k.
         &0 < w0 /\ &0 < lam /\ ~(z = &0)
         ==> gaussian_beam_intensity
             (\(a,b,z). q_parameter_amplitude A0 z a b (Cx k) w0 lam)
             (Cx x)
             (Cx y)
             z
             (Cx k) =
             norm (A0 / (ii * Cx (rayleigh_range w0 lam))) pow 2 *
             (w0 / WWW z w0 lam) pow 2 *
             exp
             (&2 *
              -- &1 *
              (k * lam) / (&2 * pi) *
              (x pow 2 + y pow 2) / WW z w0 lam pow 2)`,

MESON_TAC [(REWRITE_RULE [REAL_POW2_ABS; GSYM REAL_EXP_N] GAUSSIAN_BEAM_INTENSITY)]);;

let GAUSSIAN_BEAM_INTENSITY_ALT_2 =
   SIMP_RULE[GSYM WW_WWW] GAUSSIAN_BEAM_INTENSITY_ALT_1;;

let GAUSSIAN_BEAM_INTENSITY_MAIN = (
    `!z w0 lam A0 k.
         &0 < w0 /\ &0 < lam /\ ~(z = &0)
         ==> gaussian_beam_intensity
             (\(a,b,z). q_parameter_amplitude A0 z a b (Cx k) w0 lam)
             (Cx x) (Cx y) z  (Cx k) =
             norm (A0 / (ii * Cx (rayleigh_range w0 lam))) pow 2 *
             (w0 / WW z w0 lam) pow 2 *
             exp (--((k * lam) / pi * (x pow 2 + y pow 2) / WW z w0 lam pow 2))`,
MESON_TAC[ REWRITE_RULE[REAL_FIELD `(&2 *
              -- &1 *
              (k * lam) / (&2 * pi) *
              (x pow 2 + y pow 2) / WW z w0 lam pow 2) =  
              --((k * lam) / (pi) *
              (x pow 2 + y pow 2) / WW z w0 lam pow 2)`] GAUSSIAN_BEAM_INTENSITY_ALT_2]);;


let INV_EXP_0_AT_INFINITY = prove
 (`!x a. &0 < a ==> ((\x. exp (--a*x)) ---> &0) at_posinfinity`,
  SIMP_TAC [REALLIM_AT_POSINFINITY] THEN REPEAT STRIP_TAC THEN
  SIMP_TAC [REAL_EXP_POS_LT; REAL_FIELD `a - (&0) = a`] THEN
  SIMP_TAC [REAL_EXP_NEG; REAL_FIELD `!(a:real). (--a * x) = (--(a * x))`] THEN
  SIMP_TAC [REAL_ABS_INV] THEN
  SIMP_TAC [REAL_EXP_POS_LT; real_abs; REAL_LE_LT] THEN
  EXISTS_TAC `(log ( &1 / e)) / a + &1` THEN REPEAT STRIP_TAC THEN
  MATCH_MP_TAC REAL_LT_LINV THEN ASM_SIMP_TAC [] THEN
  ASM_SIMP_TAC [GSYM LOG_MONO_LT; REAL_LT_INV; REAL_EXP_POS_LT] THEN
  ASM_SIMP_TAC [LOG_EXP; REAL_EXP_POS_LT] THEN
  ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN IMP_REWRITE_TAC [GSYM REAL_LT_LDIV_EQ] THEN
  POP_ASSUM MP_TAC THEN
  REWRITE_TAC
    [REAL_FIELD
       `x >= log (&1 / e)/ a + &1 <=> 
                              (log (&1 / e)/ a) <= (x - &1)`] THEN
  REPEAT STRIP_TAC THEN MATCH_MP_TAC REAL_LET_TRANS THEN EXISTS_TAC `x - &1` THEN
  POP_ASSUM MP_TAC THEN ASM_SIMP_TAC [real_div; REAL_MUL_LID] THEN
  REPEAT STRIP_TAC THEN REAL_ARITH_TAC);;
